/*

    This code is part of Gf which is free software.
    See the file COPYING for more details.

    Copyright (C) 1999  T. W. Shield, shield@aem.umn.edu
    
    
    7/21/99 modified to do any number of columns

*/
#include "gfcurve.h"
#include "gfutil.h"

/**
The Constructor for an empty curve with two columns.
*/
GfCurve::GfCurve()
{
#ifdef DEBUG1
	cout << "construct gfcurve \n";
#endif	
	init();
		
}

/**
Constructor for curves with more than two columns.
<pre>
	number of columns only affects the amount of data 
	storage.  The is one column set as X and one set as Y
	for plotting purposes.  To plot multiple curves you need 
	multiple curves, not many columns
</pre>
*/
GfCurve::GfCurve(int n)
{
#ifdef DEBUG1
	cout << "construct gfcurve \n";
#endif	

	init();
	if(n >= 2)
		numcols = n;
		
}

void GfCurve::init()
{

	data = NULL;

	numcols = 2; // minimum value of numcols
	xcol = 0;
	ycol = 1;
	
	invec = NULL;
	numpoints = 0;
	
	linetype = LT_SOLID;
	linewidth = LW_DEFAULT;
	c_color = BLACK;
	patlen = 4.;
	
		
}

void GfCurve::storage(int size)
{
/*
	numcols must be set correctly before calling this
	
*/
	if(size > 0)
	{
		data = (float *)realloc(data,numcols*size*sizeof(float));
		if( data == NULL )
		{
			printf("out of memory in gfcurve\n");
			exit(-1);
		}

		invec = (int *)realloc(invec,size*sizeof(int));
		if( invec == NULL )
		{
			printf("out of memory in gfcurve\n");
			exit(-1);
		}
	}
	else
	{
		free(data);
		free(invec);
		
		data = NULL;
		invec = NULL;
	}
}

/*
Copy Constructor.
*/
GfCurve::GfCurve(const GfCurve &source) // copy constructor
{
int i;
float *pt;

	init();

	numpoints = source.numPoints();
	
	numcols = source.numColumns();
	
	storage(numpoints);
	
	xcol = source.xColumn();
	ycol = source.yColumn();
	
	pt = mallocRow();
	
	for(i=0;i<numpoints;i++)
	{
		source.row(i,pt);
		setRow(i,pt);
		invec[i] = source.pointInvec(i);
	}
	
	free(pt);
	
	patlen=source.patternLength();	
	linetype=source.lineType();	
	c_color=source.color();	
	linewidth=source.lineWidth();
}

/*
Copies a segment of a curve into the current curve.
<pre>

	call with start = SEG_WHOLE_CURVE to copy whole curve
	
	call with start = SEG_LAST_LINE to copy just last 2 points
	
	to copy one point call with start == end
	
	returns:
		GF_OK on success
		GF_BAD on failure
</pre>
*/
int GfCurve::copySegment(const GfCurve *source,int start,int end)  
{
int i;
float *pt;

	numpoints = source->numPoints();
	
	if(start == SEG_WHOLE_CURVE)
	{	
		start = 0;
		end = numpoints-1;
	}
	
	if(start == SEG_LAST_LINE)
	{	
		start = numpoints-2;
		end = numpoints-1;
	}
			
	if(start == SEG_ONE_POINT) // or just pass in start=end
	{	
		start = end;
	}		
	
	if(start < 0 || start>= numpoints || end < start || end >= numpoints)
	{
		numpoints = 0;
		return(GF_BAD);
	}
	
	numpoints = end - start + 1;
	
	numcols = source->numColumns();
	
	storage(numpoints);
		
	xcol = source->xColumn();
	ycol = source->yColumn();
	
	pt = mallocRow();
	
	for(i=start;i<=end;i++)
	{
		source->row(i,pt);
		setRow(i-start,pt);
		invec[i-start] = source->pointInvec(i);
	}
	
	patlen=source->patternLength();	
	linetype=source->lineType();	
	c_color=source->color();	
	linewidth=source->lineWidth();
	
	if(invec[0] != 0)
	{	
		if(invec[0] != 1)
		{
			row(0,pt);
			insertRow(0,invec[0],pt);
		}
		invec[0]=0;
	}
		
	free(pt);
	
	return(GF_OK);
}

GfCurve::GfCurve(const GfCurve *source) // copy constructor from pointer
{
int i;
float *pt;

	init();

	numpoints = source->numPoints();
	
	numcols = source->numColumns();
	
	storage(numpoints);
	
	xcol = source->xColumn();
	ycol = source->yColumn();
	
	pt = mallocRow();
	
	for(i=0;i<numpoints;i++)
	{
		source->row(i,pt);
		setRow(i,pt);
		invec[i] = source->pointInvec(i);
	}
	
	free(pt);
	
	patlen=source->patternLength();	
	linetype=source->lineType();	
	c_color=source->color();	
	linewidth=source->lineWidth();
}

GfCurve::~GfCurve()
{
	free(data);
	free(invec);
}


void GfCurve::print(OSTREAM *os)
{
int i;
	//format(os); don't need this in clibboard -- only use of this
	// function I think ;)
	
	for(i=0;i<numpoints;i++)
	{
		*os <<invec[i]<<"\t"<< xp(i) << "\t"<<yp(i)<<"\n";
	}

} 

void GfCurve::format(OSTREAM *os)
{
	*os<<"C\tpatternlength: "<<patlen<<"\n";
	*os<<"C\tlinetype: "<<linetype<<"\n";
	*os<<"C\tcolor: "<<c_color<<"\n";
	*os<<"C\tlinewidth: "<<linewidth<<"\n";
}

int GfCurve::readFormat(QFile *in,char *line)
{

	in->readLine(line,MAX_LINE_LEN-1);
	if(in->atEnd())
	{
		return(GF_BAD);
	}

	if(getfloatval("patternlength",line,&patlen) != GF_OK)
		return(GF_BAD);
		
	in->readLine(line,MAX_LINE_LEN-1);
	if(in->atEnd())
	{
		return(GF_BAD);
	}
	
	if(getintval("linetype",line,&linetype) != GF_OK)
		return(GF_BAD);
		
	in->readLine(line,MAX_LINE_LEN-1);
	if(in->atEnd())
	{
		return(GF_BAD);
	}
	
	if(getintval("color",line,&c_color) != GF_OK)
		return(GF_BAD);
		
	in->readLine(line,MAX_LINE_LEN-1);
	if(in->atEnd())
	{
		return(GF_BAD);
	}
	
	if(getintval("linewidth",line,&linewidth) != GF_OK)
		return(GF_BAD);
		
	in->readLine(line,MAX_LINE_LEN-1);
	if(in->atEnd())
	{
		return(GF_BAD);
	}
	
	return(GF_OK);

}

OSTREAM &operator<<(OSTREAM &os, GfCurve *m)
{
	m->print(&os);
	return(os);
}
OSTREAM &operator<<(OSTREAM &os, GfCurve &m)
{
	m.print(&os);
	return(os);
}

void GfCurve::operator=(const GfCurve &source)
{
int i;
float *pt;

	numpoints = source.numPoints();
	
	numcols = source.numColumns();
	
	storage(numpoints);
	
	xcol = source.xColumn();
	ycol = source.yColumn();
	
	pt = mallocRow();
	
	for(i=0;i<numpoints;i++)
	{
		source.row(i,pt);
		setRow(i,pt);
		invec[i] = source.pointInvec(i);
	}
	
	free(pt);
	
	patlen=source.patternLength();	
	linetype=source.lineType();	
	c_color=source.color();	
	linewidth=source.lineWidth();
}

int GfCurve::setPoint(int i,int iv,float x,float y)
{

	if( i < numpoints && i > -1)
	{
		setXp(i,x);
		setYp(i,y);
		invec[i]=iv;
		
		if(i == 0)
		{
			if(iv != 0 )
			{
				if(iv != 1)
				{
					insertPoint(0,iv,x,y);
				}
				invec[0] = 0;
			}
		}
		
		return(GF_OK);
	}
	else
		return(GF_BAD);
}
	
int GfCurve::appendPoint(int iv,float x,float y)
{
	if(numpoints == 0 && iv != 0) // force first point to have iv==0
	{
		if(iv == 1)
		{
			iv=0; // if iv == 1 same effect as 0
		}
		else
		{
			storage(numpoints+1);
	
			zeroRow(numpoints);
	
			setXp(numpoints,x); // duplicate point with iv=0
			setYp(numpoints,y);
			invec[numpoints] = 0;
	
			numpoints++;		
		
		
		}
	}

	storage(numpoints+1);
	
	zeroRow(numpoints);
	
	setXp(numpoints,x); 
	setYp(numpoints,y);
	invec[numpoints] = iv;
	
	numpoints++;
	
	return(GF_OK);
}
	
int GfCurve::insertPoint(int after, int iv,float x,float y)
{
int j;
float *pt;
	if( after < numpoints && after > -1)
	{	
		storage(numpoints+1);
	
		if( after < (numpoints - 1) )
		{
			pt = mallocRow();
			
			for(j=(numpoints-1);j>after;j--)
			{
				row(j,pt);
				setRow(j+1,pt);
				invec[j+1] = invec[j];
			}
			
			free(pt);
		}
		
		zeroRow(after+1);
		setXp(after+1,x);
		setYp(after+1,y);
		invec[after+1]=iv;
		
		numpoints++;
		return(GF_OK);
			
	}
	else
		return(GF_BAD);
		
}

int GfCurve::removePoint(int i)
{
int j;
float *pt;
	if( i < numpoints && i > -1)
	{
	
		if( i < (numpoints - 1) )
		{
			pt = mallocRow();
			
			for(j=i;j<(numpoints-1);j++)
			{
				row(j+1,pt);
				setRow(j,pt);
				invec[j] = invec[j+1];
			}
			
			free(pt);
		}
		
		storage(numpoints-1);
		numpoints--;
		return(GF_OK);
			
	}
	else
		return(GF_BAD);
		
}
	
	
float GfCurve::pointX(int i) const 
{
	return(xp(i));
}

float GfCurve::pointY(int i) const 
{
	return(yp(i));
}


int GfCurve::pointInvec(int i) const 
{
	if( i < numpoints && i > -1)
		return(invec[i]);
	else
		return(0);
}
	
int GfCurve::numPoints() const 
{
	return(numpoints);
}
	
int GfCurve::setPatternLength(float t)
{
	if( t > 0 )
	{
		patlen = t;
		return(GF_OK);
	}
	else
	{
		return(GF_BAD);
	}
}

float GfCurve::patternLength() const 
{
	return(patlen);
}

int GfCurve::setColor(int t)
{
	if( t >= 0 && t < NUM_PSCOLORS)
	{
		c_color = t;
		return(GF_OK);
	}
	else
	{
		return(GF_BAD);
	}
}

int GfCurve::color() const 
{
	return(c_color);
}

int GfCurve::setLineType(int t)
{
	if( t >= 0 && t < NUM_PSLINETYPES)
	{
		linetype = t;
		return(GF_OK);
	}
	else
	{
		return(GF_BAD);
	}
}

int GfCurve::lineType() const 
{
	return(linetype);
}

int GfCurve::setLineWidth(int t)
{
	if( t >= 0 && t < NUM_PSLINEWIDTHS)
	{
		linewidth = t;
		return(GF_OK);
	}
	else
	{
		return(GF_BAD);
	}
}

int GfCurve::lineWidth() const 
{
	return(linewidth);
}

void GfCurve::getMaxMin(float *xmax,float *xmin,float *ymax,float *ymin) const 
{
int i;
	*xmax = -1.e36;
	*xmin = 1.e36;
	*ymax = *xmax;
	*ymin = *xmin;

	
	for(i=0;i<numpoints;i++)
	{
		if(xp(i) > *xmax) *xmax = xp(i);
		if(xp(i) < *xmin) *xmin = xp(i);
		
		if(yp(i) > *ymax) *ymax = yp(i);
		if(yp(i) < *ymin) *ymin = yp(i);
	}

}

void GfCurve::getAverages(float *xave,float *yave) const 
{
int i;
	*xave = 0.;
	*yave = 0.;

	
	for(i=0;i<numpoints;i++)
	{
		*xave = *xave + xp(i);
		*yave = *yave + yp(i);
	}
	
	if(numpoints > 0)
	{
		*xave=*xave/numpoints;
		*yave=*yave/numpoints;
	}

}

int GfCurve::linearFit(float *slope,float *intercept,int start,int end)
{

        int n,i;
        float sumx,sumy,sumxx,sumxy,denom;

        sumxx = 0.;
        sumy = 0.;
        sumxy = 0.;
        sumx = 0.;
	
	if(start == SEG_WHOLE_CURVE)
	{	
		start = 0;
		end = numpoints-1;
	}
	
	if(start < 0 || start >= numpoints || end < start || end >= numpoints)
	{
		return(GF_BAD);
	}
	
        for(i=start;i<=end;i++)
        {
		sumxx = sumxx + xp(i) * xp(i);
                sumxy = sumxy + xp(i) * yp(i);
                sumy = sumy + yp(i);
                sumx = sumx + xp(i);
        }

        n = end - start + 1;
        denom = n * sumxx - sumx * sumx;

        if(denom != 0. )
        {
               *slope = (n * sumxy - sumx * sumy) / denom ;
               *intercept = (sumy * sumxx - sumx * sumxy) / denom ;
	       return(GF_OK);
	}
	
	return(GF_BAD);
	
}

void GfCurve::linearTransform(float a,float b,float c,float d,float e,float f)
{
int i;
float x,y;

	for(i = 0;i < numpoints;i++)
	{
		x = xp(i);
		y = yp(i);
		setXp(i, a * x + b * y + c);
		setYp(i, d * x + e * y + f);
	}
}

int GfCurve::reduceNumberPoints(int step)
{
/*
	step = 1 keeps all points
	step = 2 keeps every 2nd point, etc.
	
	first and last points are always kept
	
*/
int i,k,n;

	n = (numpoints+1)/step;
	
	if(n < 10)
	{
		return(GF_BAD);
	}

	k = step;

	for(i=1;i<n;i++)
	{
		setXp(i,xp(k));
		setYp(i,yp(k));
		invec[i] = invec[k];
		k=k+step;
		if(k >= numpoints)
		{
			n = i+1;
			break;
		}
	}
	
	if((n*step) != (numpoints+1))
	{
		setXp(n,xp(numpoints-1));
		setYp(n,yp(numpoints-1));
		invec[n] = invec[numpoints-1];
		n++;
	}

	numpoints = n;
	
	storage(numpoints);
	
	return(GF_OK);
}

int GfCurve::smooth(int half_window)
{
/*
	does an inplace moving window of (2*half_window+1) width
	average on the data.  Does not smooth the half_window points
	at either end.
	
	could also do other windows than square.
*/
int i,j;
float *xprev,*yprev,n,xnew,ynew;

	if(half_window < 1)
	{
		return(GF_BAD);
	}
	
	if(numpoints-half_window < half_window)
	{
		return(GF_BAD);
	}
	
	xprev = (float *)malloc((half_window+1)*sizeof(float));
	if(xprev == NULL)
	{
		cout << "out of memory in smooth\n";
		exit(-1);
	}
	
	yprev = (float *)malloc((half_window+1)*sizeof(float));
	if(yprev == NULL)
	{
		cout << "out of memory in smooth\n";
		exit(-1);
	}
	
	n =(float)( 2*half_window+1);
	
	
	for(i=0;i<=half_window;i++)
	{
		xprev[i]=xp(i);
		yprev[i]=yp(i);
	}			
	
	for(i=half_window;i<numpoints-half_window;i++)
	{
		xnew = 0.;
		ynew = 0.;
		
		for(j=0;j<=half_window;j++)
		{
			xnew = xnew + xprev[j];
			ynew = ynew + yprev[j];
		}
		
		for(j=1;j<=half_window;j++)
		{
			xnew = xnew + xp(i+j);
			ynew = ynew + yp(i+j);
		}
	
		xnew = xnew/n;
		ynew = ynew/n;
		
		for(j=0;j<half_window;j++)
		{
			xprev[j]=xprev[j+1];
			yprev[j]=yprev[j+1];
		}
		xprev[half_window] = xp(i);
		yprev[half_window] = yp(i);
		
		setXp(i,xnew);
		setYp(i,ynew);
	
	}

	free(xprev);
	free(yprev);
	
	return(GF_OK);
}

/* *****************************************************

	multi-column support functions
	
*******************************************************/

int GfCurve::numColumns()  const // use constructor to set numcols
{
	return(numcols);
}

int GfCurve::xColumn() const 
{
	return(xcol);
}

int GfCurve::yColumn() const 
{
	return(ycol);
}

int GfCurve::setXColumn(int c)
{
	if(c > -1 && c < numcols)
	{
		xcol = c;
		return(GF_OK);
	}
	else
		return(GF_BAD);
}

int GfCurve::setYColumn(int c)
{
	if(c > -1 && c < numcols)
	{
		ycol = c;
		return(GF_OK);
	}
	else
		return(GF_BAD);
}

float GfCurve::xp(int i) const 
{
	if(i< 0 || i> numpoints)
		return(0.0);
		
	return(data[i*numcols+xcol]);
}
	
float GfCurve::yp(int i) const 
{
	if(i< 0 || i> numpoints)
		return(0.0);
		
	return(data[i*numcols+ycol]);
}	

int GfCurve::setXp(int i,float x)
{
	if(i< 0 || i> numpoints)
		return(GF_BAD);
		
	data[i*numcols+xcol]=x;
	
	return(GF_OK);
}
	
int GfCurve::setYp(int i,float y)
{
	if(i< 0 || i> numpoints)
		return(GF_BAD);
		
	data[i*numcols+ycol]=y;
	
	return(GF_OK);
}
	
int GfCurve::row(int i,float *pt) const 
{
int j;
	if(i< 0 || i> numpoints)
		return(GF_BAD);
		
	for(j=0;j<numcols;j++)
	{
		pt[j] = data[i*numcols+j];
	}
	
	return(GF_OK);
}

int GfCurve::setRow(int i,float *pt)
{
int j;
	if(i< 0 || i> numpoints)
		return(GF_BAD);
		
	for(j=0;j<numcols;j++)
	{
		data[i*numcols+j] = pt[j];
	}
	
	return(GF_OK);
}

float *GfCurve::mallocRow()
{
float *pt;

 	pt = (float *)malloc(numcols*sizeof(float));
	
	if(pt == NULL)
	{
		cout <<"out of memory in GfCurve::mallocRow\n";
		exit(-1);
	}
	
	return(pt);
}

int GfCurve::zeroRow(int i)
{
int j;
	if(i< 0 || i> numpoints)
		return(GF_BAD);
		
	for(j=0;j<numcols;j++)
	{
		data[i*numcols+j] = 0.0;
	}
	
	return(GF_OK);
}

int GfCurve::appendRow(int iv,float *row)
{
	if(numpoints == 0 && iv != 0) // force first point to have iv==0
	{
		if(iv == 1)
		{
			iv=0; // if iv == 1 same effect as 0
		}
		else
		{
			storage(numpoints+1);
	
			setRow(numpoints,row); // duplicate point with iv=0
			
			invec[numpoints] = 0;
	
			numpoints++;		
		
		
		}
	}

	storage(numpoints+1);
	
	setRow(numpoints,row); 

	invec[numpoints] = iv;
	
	numpoints++;
	
	return(GF_OK);
}
	
int GfCurve::insertRow(int after, int iv,float *r)
{
int j;
float *pt;
	if( after < numpoints && after > -1)
	{	
		storage(numpoints+1);
	
		if( after < (numpoints - 1) )
		{
			pt = mallocRow();
			
			for(j=(numpoints-1);j>after;j--)
			{
				row(j,pt);
				setRow(j+1,pt);
				invec[j+1] = invec[j];
			}
			
			free(pt);
		}
		
		setRow(after+1,r);
		invec[after+1]=iv;
		
		numpoints++;
		return(GF_OK);
			
	}
	else
		return(GF_BAD);
		
}

int GfCurve::removeRow(int i)
{
	return(removePoint(i));
}

int GfCurve::sprintRow(char *text,int i,int vlen,int dlen,int line_fmt,char delim)
{
char ivfmt[10],datafmt[10],temp[100];
int j;

	if( i< 0 || i>= numpoints )
		return(GF_BAD);
		
	if(dlen < 8) return(GF_BAD);

        if(line_fmt == FMT_FIXED)
        {
              if(vlen > 0)
              {
                    sprintf(ivfmt,"%%%ii",vlen);
              }
              sprintf(datafmt,"%%+%i.%iE",dlen,dlen-7);
        }
	
	strcpy(text,"");
	
	if(vlen > 0)
	{
		if(line_fmt == FMT_FIXED)
		{
			sprintf(text,ivfmt,invec[i]);
		}
		else
		{
			sprintf(text,"%i",invec[i]);
		}
	}
	
	// should we output xcol,ycol, rest... ?
	// this ignores xcol and ycol
	
	for(j=0;j<numcols;j++)
	{
		if(line_fmt == FMT_FIXED)
		{
			sprintf(temp,datafmt,data[i*numcols+j]);
#ifndef UNIX			
			if(strlen(temp) > dlen) 
				winFixExpFmt(temp);
#endif
		}
		else
		{
			if(vlen > 0 || j > 0)
				sprintf(temp,"%c%g",delim,data[i*numcols+j]);
			else
				sprintf(temp,"%g",data[i*numcols+j]);
		}
		strcat(text,temp);
	}
	
	strcat(text,"\n");
	
	return(GF_OK);
}
